from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from edc_utils import get_utcnow

from ..exceptions import AllocationError

if TYPE_CHECKING:
    from ..models import StockRequest


def allocate_stock(
    stock_request: StockRequest,
    allocation_data: dict[str, str],
    allocated_by: str,
    user_created: str = None,
    created: datetime = None,
) -> tuple[int, int]:
    stock_model_cls = django_apps.get_model("edc_pharmacy.stock")
    allocation_model_cls = django_apps.get_model("edc_pharmacy.allocation")
    registered_subject_model_cls = django_apps.get_model("edc_registration.registeredsubject")
    allocated, skipped = 0, 0
    stock_objs = []
    for code, subject_identifier in allocation_data.items():
        rs_obj = registered_subject_model_cls.objects.get(
            subject_identifier=subject_identifier
        )
        stock_request_item = stock_request.stockrequestitem_set.filter(
            registered_subject=rs_obj,
            allocation__isnull=True,
        ).first()
        if not stock_request_item:
            skipped += 1
            continue
        try:
            stock_obj = stock_model_cls.objects.get(
                code=code,
                container__may_request_as=True,
                confirmed=True,
                allocation__isnull=True,
            )
        except ObjectDoesNotExist:
            skipped += 1
        else:
            allocation = allocation_model_cls.objects.create(
                stock_request_item=stock_request_item,
                registered_subject=rs_obj,
                allocation_datetime=get_utcnow(),
                allocated_by=allocated_by,
                user_created=user_created,
                created=created,
            )
            if (
                stock_model_cls.objects.get(code=code).product.assignment
                != allocation.get_assignment()
            ):
                allocation.delete()
                raise AllocationError(
                    "Assignment mismatch. Stock must match subject assignment. "
                    "Allocation abandoned."
                )

            stock_obj.allocation = allocation
            stock_obj.user_modified = user_created
            stock_obj.modified = created
            stock_objs.append(stock_obj)
    if stock_objs:
        for obj in stock_objs:
            obj.save()
            allocated += 1
    return allocated, skipped


__all__ = ["allocate_stock"]



# Generated by Django 5.1.2 on 2024-11-26 14:21

import _socket
import django.core.validators
import django.db.models.deletion
import django_audit_fields.fields.hostname_modification_field
import django_audit_fields.fields.userfield
import django_audit_fields.models.audit_model_mixin
import django_revision.revision_field
import edc_pharmacy.models.stock.location
import simple_history.models
import uuid
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0052_formulation_imp_formulation_imp_description_and_more"),
        ("sites", "0002_alter_domain_unique"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AlterModelManagers(
            name="location",
            managers=[
                ("objects", edc_pharmacy.models.stock.location.Manager()),
            ],
        ),
        migrations.AlterField(
            model_name="historicalstock",
            name="lot",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="edc_pharmacy.lot",
                verbose_name="Batch",
            ),
        ),
        migrations.AlterField(
            model_name="stock",
            name="lot",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="edc_pharmacy.lot",
                verbose_name="Batch",
            ),
        ),
        migrations.CreateModel(
            name="HistoricalLocation",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True, default=django_audit_fields.models.audit_model_mixin.utcnow
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True, default=django_audit_fields.models.audit_model_mixin.utcnow
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                        verbose_name="Hostname created",
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                        verbose_name="Hostname modified",
                    ),
                ),
                (
                    "device_created",
                    models.CharField(blank=True, max_length=10, verbose_name="Device created"),
                ),
                (
                    "device_modified",
                    models.CharField(
                        blank=True, max_length=10, verbose_name="Device modified"
                    ),
                ),
                (
                    "locale_created",
                    models.CharField(
                        blank=True,
                        help_text="Auto-updated by Modeladmin",
                        max_length=10,
                        null=True,
                        verbose_name="Locale created",
                    ),
                ),
                (
                    "locale_modified",
                    models.CharField(
                        blank=True,
                        help_text="Auto-updated by Modeladmin",
                        max_length=10,
                        null=True,
                        verbose_name="Locale modified",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        db_index=True,
                        help_text="This is the stored value, required",
                        max_length=250,
                        verbose_name="Stored value",
                    ),
                ),
                (
                    "plural_name",
                    models.CharField(max_length=250, null=True, verbose_name="Plural name"),
                ),
                (
                    "display_index",
                    models.IntegerField(
                        default=0,
                        help_text="Index to control display order if not alphabetical, not required",
                        verbose_name="display index",
                    ),
                ),
                (
                    "field_name",
                    models.CharField(
                        blank=True,
                        editable=False,
                        help_text="Not required",
                        max_length=25,
                        null=True,
                    ),
                ),
                ("extra_value", models.CharField(max_length=250, null=True)),
                ("version", models.CharField(default="1.0", editable=False, max_length=35)),
                ("id", models.IntegerField(blank=True, db_index=True)),
                (
                    "display_name",
                    models.CharField(
                        blank=True,
                        db_index=True,
                        help_text="(suggest 40 characters max.)",
                        max_length=250,
                        null=True,
                        verbose_name="Name",
                    ),
                ),
                ("contact_name", models.CharField(blank=True, max_length=150, null=True)),
                (
                    "contact_tel",
                    models.CharField(
                        blank=True,
                        max_length=150,
                        null=True,
                        validators=[django.core.validators.RegexValidator("[0-9]{1,15}")],
                    ),
                ),
                ("contact_email", models.EmailField(blank=True, max_length=150, null=True)),
                (
                    "history_id",
                    models.UUIDField(
                        default=uuid.uuid4, editable=False, primary_key=True, serialize=False
                    ),
                ),
                ("history_date", models.DateTimeField(db_index=True)),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "site",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="sites.site",
                    ),
                ),
            ],
            options={
                "verbose_name": "historical Location",
                "verbose_name_plural": "historical Locations",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": ("history_date", "history_id"),
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
    ]



from dateutil.relativedelta import relativedelta
from django.test import TestCase
from edc_registration.models import RegisteredSubject
from edc_utils import get_utcnow

from edc_pharmacy.models import (
    DosageGuideline,
    Formulation,
    FormulationType,
    FrequencyUnits,
    Medication,
    Route,
    Rx,
    Units,
)
from edc_pharmacy.refill import RefillCreator


class TestDispense(TestCase):
    def setUp(self):
        self.subject_identifier = "12345"

        RegisteredSubject.objects.create(subject_identifier="12345")

        self.medication = Medication.objects.create(
            name="flucytosine",
            display_name="Flucytosine",
        )

        self.formulation = Formulation.objects.create(
            medication=self.medication,
            strength=500,
            units=Units.objects.get(name="mg"),
            route=Route.objects.get(display_name="Oral"),
            formulation_type=FormulationType.objects.get(display_name__iexact="Tablet"),
        )

        self.dosage_guideline = DosageGuideline.objects.create(
            medication=self.medication,
            dose_per_kg=100,
            dose_units=Units.objects.get(name="mg"),
            frequency=1,
            frequency_units=FrequencyUnits.objects.get(name="day"),
        )

        self.rx = Rx.objects.create(
            subject_identifier=self.subject_identifier,
            weight_in_kgs=40,
            report_datetime=get_utcnow(),
        )
        self.rx.medications.add(self.medication)

    def test_dispense(self):
        refill_creator = RefillCreator(
            subject_identifier=self.subject_identifier,
            refill_start_datetime=get_utcnow(),
            refill_end_datetime=get_utcnow() + relativedelta(days=7),
            dosage_guideline=self.dosage_guideline,
            formulation=self.formulation,
        )
        self.assertEqual(refill_creator.rx_refill.total, 56.0)
        self.assertEqual(refill_creator.rx_refill.remaining, 56.0)



# Generated by Django 3.2.13 on 2022-08-26 01:06

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_pharmacy", "0010_auto_20220708_0813"),
    ]

    operations = [
        migrations.AlterField(
            model_name="historicalrx",
            name="action_identifier",
            field=models.CharField(blank=True, db_index=True, max_length=50, null=True),
        ),
        migrations.AlterField(
            model_name="rx",
            name="action_identifier",
            field=models.CharField(blank=True, max_length=50, null=True, unique=True),
        ),
    ]



from django.utils.translation import gettext as _
from edc_pdf_reports import NumberedCanvas as BaseNumberedCanvas
from edc_pdf_reports import Report
from edc_pdf_reports.flowables import CheckboxFlowable, TextFieldFlowable
from edc_protocol.research_protocol_config import ResearchProtocolConfig
from edc_utils.date import to_local
from reportlab.graphics.barcode import code128
from reportlab.lib import colors
from reportlab.lib.enums import TA_CENTER, TA_LEFT, TA_RIGHT
from reportlab.lib.pagesizes import A4
from reportlab.lib.styles import ParagraphStyle
from reportlab.lib.units import cm, mm
from reportlab.pdfbase.pdfmetrics import stringWidth
from reportlab.platypus import Paragraph, SimpleDocTemplate, Spacer, Table, TableStyle

from ..models import StockTransfer


class NumberedCanvas(BaseNumberedCanvas):
    footer_row_height = 60


class ManifestReport(Report):

    def __init__(self, stock_transfer: StockTransfer = None, **kwargs):
        self.stock_transfer = stock_transfer
        self.protocol_name = ResearchProtocolConfig().protocol_title
        super().__init__(**kwargs)

    def draw_header(self, canvas, doc):
        width, height = A4
        canvas.setFontSize(6)
        text_width = stringWidth(self.protocol_name, "Helvetica", 6)
        canvas.drawRightString(width - text_width, height - 20, self.protocol_name.upper())
        canvas.drawString(
            40,
            height - 30,
            (
                _("Stock Transfer Manifest: %(transfer_identifier)s")
                % {"transfer_identifier": self.stock_transfer.transfer_identifier}
            ).upper(),
        )

    @property
    def queryset(self):
        return self.stock_transfer.stocktransferitem_set.all().order_by(
            "stock__allocation__registered_subject__subject_identifier"
        )

    def get_report_story(self, document_template: SimpleDocTemplate = None, **kwargs):
        story = []

        data = [
            [
                Paragraph(
                    _("Stock Transfer Manifest").upper(),
                    ParagraphStyle(
                        "Title",
                        fontSize=10,
                        spaceAfter=0,
                        alignment=TA_LEFT,
                        fontName="Helvetica-Bold",
                    ),
                ),
                Paragraph(
                    self.protocol_name.upper(),
                    ParagraphStyle(
                        "Title",
                        fontSize=10,
                        spaceAfter=0,
                        alignment=TA_RIGHT,
                        fontName="Helvetica-Bold",
                    ),
                ),
            ],
        ]
        table = Table(data)
        story.append(table)
        story.append(Spacer(0.1 * cm, 0.5 * cm))

        bold_left_style = ParagraphStyle(
            name="line_data_medium", alignment=TA_LEFT, fontSize=8, fontName="Helvetica-Bold"
        )
        bold_right_style = ParagraphStyle(
            name="line_data_medium",
            alignment=TA_RIGHT,
            fontSize=8,
            fontName="Helvetica-Bold",
        )
        left_style = ParagraphStyle(name="line_data_medium", alignment=TA_LEFT, fontSize=8)
        right_style = ParagraphStyle(name="line_data_medium", alignment=TA_RIGHT, fontSize=8)
        from_location = self.stock_transfer.from_location.display_name
        contact_name = self.stock_transfer.from_location.contact_name or ""
        tel = self.stock_transfer.from_location.contact_tel or ""
        email = self.stock_transfer.from_location.contact_email or ""
        timestamp = to_local(self.stock_transfer.transfer_datetime).strftime("%Y-%m-%d")
        data = [
            [
                Paragraph(_("Reference:"), bold_left_style),
                Paragraph(self.stock_transfer.transfer_identifier, left_style),
                Paragraph(_("Contact:"), bold_right_style),
            ],
            [
                Paragraph(_("Date:"), bold_left_style),
                Paragraph(timestamp, left_style),
                Paragraph(contact_name, right_style),
            ],
            [
                Paragraph(_("From:"), bold_left_style),
                Paragraph(from_location, left_style),
                Paragraph(email, right_style),
            ],
            [
                Paragraph(_("To:"), bold_left_style),
                Paragraph(self.stock_transfer.to_location.display_name, left_style),
                Paragraph(tel, right_style),
            ],
            [
                Paragraph("Items:", bold_left_style),
                Paragraph(str(self.queryset.count()), left_style),
                Paragraph("", right_style),
            ],
        ]
        text_width1 = stringWidth(_("Reference"), "Helvetica", 10)
        table = Table(
            data, colWidths=(text_width1 * 1.5, None, None), rowHeights=(10, 10, 10, 10, 10)
        )
        story.append(table)

        story.append(self.stock_transfer_items_as_table)

        story.append(Spacer(0.1 * cm, 0.5 * cm))

        story.append(self.signature_line_as_table)

        story.append(Spacer(0.1 * cm, 0.5 * cm))

        story.append(self.comment_box_as_table)

        return story

    @property
    def stock_transfer_items_as_table(self) -> Table:

        style = ParagraphStyle(
            name="line_data_medium",
            alignment=TA_CENTER,
            fontSize=8,
            textColor=colors.black,
            fontName="Helvetica-Bold",
        )
        data = [
            [
                Paragraph("", style),
                Paragraph("#", style),
                Paragraph(_("Subject"), style),
                Paragraph(_("Code"), style),
                Paragraph(_("Barcode"), style),
                Paragraph(_("Formulation"), style),
                Paragraph(_("Pack"), style),
            ]
        ]
        for index, stock_transfer_item in enumerate(self.queryset):
            barcode = code128.Code128(
                stock_transfer_item.stock.code, barHeight=5 * mm, barWidth=0.7, gap=1.7
            )
            subject_identifier = (
                stock_transfer_item.stock.allocation.registered_subject.subject_identifier
            )
            formulation = stock_transfer_item.stock.product.formulation
            description = f"{formulation.imp_description} "
            style = ParagraphStyle(
                name="line_data", alignment=TA_CENTER, fontSize=8, leading=10
            )
            style_xsmall = ParagraphStyle(
                name="line_data", alignment=TA_CENTER, fontSize=6, leading=8
            )
            data.append(
                [
                    CheckboxFlowable(name=f"checkbox_{index}"),
                    Paragraph(str(index + 1), style),
                    Paragraph(subject_identifier, style),
                    Paragraph(
                        (
                            f"{stock_transfer_item.stock.code[:3]}-"
                            f"{stock_transfer_item.stock.code[3:]}"
                        ),
                        style,
                    ),
                    barcode,
                    Paragraph(description, style_xsmall),
                    Paragraph(str(stock_transfer_item.stock.container), style),
                ]
            )

        table = Table(
            data,
            colWidths=(0.5 * cm, 1 * cm, None, None, None, None, None),
        )
        table.setStyle(
            TableStyle(
                [
                    ("BACKGROUND", (0, 0), (-1, 0), colors.white),
                    ("TEXTCOLOR", (0, 0), (-1, 0), colors.black),
                    ("ALIGN", (0, 0), (-1, -1), "CENTER"),
                    ("FONTNAME", (0, 0), (-1, 0), "Helvetica-Bold"),
                    ("BOTTOMPADDING", (0, 0), (-1, 0), 0),
                    ("BACKGROUND", (0, 1), (-1, -1), colors.white),
                    ("GRID", (0, 0), (-1, -1), 1, colors.black),
                    ("VALIGN", (0, 0), (-1, -1), "MIDDLE"),
                ]
            )
        )
        return table

    @property
    def signature_line_as_table(self) -> Table:
        style = ParagraphStyle(
            name="line_data_medium", alignment=TA_LEFT, fontSize=8, leading=8
        )
        textfield_width = stringWidth("_________________________", "Helvetica", 8)
        data = [
            [
                TextFieldFlowable(
                    name="issued_by_signature",
                    value="",
                    width=textfield_width,
                    height=10,
                    borderWidth=0.5,
                    fontSize=8,
                ),
                TextFieldFlowable(
                    name="received_by_signature",
                    value="",
                    width=textfield_width,
                    height=10,
                    borderWidth=0.5,
                    fontSize=8,
                ),
                TextFieldFlowable(
                    name="received_by",
                    value="",
                    width=textfield_width,
                    height=10,
                    borderWidth=0.5,
                    fontSize=8,
                ),
                TextFieldFlowable(
                    name="received_count",
                    value="",
                    width=textfield_width / 2,
                    height=10,
                    borderWidth=0.5,
                    fontSize=8,
                ),
            ],
            [
                Paragraph(_("Issued by: signature /date"), style=style),
                Paragraph(_("Received by: signature / date"), style=style),
                Paragraph(_("Received by: printed name"), style=style),
                Paragraph(_("Received count"), style=style),
            ],
        ]
        table = Table(data, colWidths=(None, None, None, None), rowHeights=(10, 10))
        return table

    @property
    def comment_box_as_table(self) -> Table:
        style = ParagraphStyle(
            name="line_data_medium",
            fontSize=8,
            textColor=colors.black,
            fontName="Helvetica-Bold",
        )
        data = [[Paragraph(_("Comment:"), style)]]
        table = Table(data, rowHeights=(75,))
        table.setStyle(
            TableStyle(
                [
                    ("BACKGROUND", (0, 0), (-1, 0), colors.white),
                    ("TEXTCOLOR", (0, 0), (-1, 0), colors.black),
                    ("ALIGN", (0, 0), (-1, -1), "CENTER"),
                    ("FONTNAME", (0, 0), (-1, 0), "Helvetica-Bold"),
                    ("BOTTOMPADDING", (0, 0), (-1, 0), 0),
                    ("BACKGROUND", (0, 1), (-1, -1), colors.white),
                    ("GRID", (0, 0), (-1, -1), 1, colors.black),
                    ("VALIGN", (0, 0), (-1, -1), "TOP"),
                ]
            )
        )
        return table



from datetime import date

from django.db import models
from edc_consent.managers import ConsentObjectsByCdefManager, CurrentSiteByCdefManager
from edc_crf.model_mixins import CrfModelMixin
from edc_identifier.managers import SubjectIdentifierManager
from edc_identifier.model_mixins import NonUniqueSubjectIdentifierFieldMixin
from edc_model.models import BaseUuidModel
from edc_registration.model_mixins import UpdatesOrCreatesRegistrationModelMixin
from edc_screening.model_mixins import ScreeningModelMixin
from edc_sites.model_mixins import SiteModelMixin
from edc_utils import get_utcnow
from edc_visit_schedule.model_mixins import OffScheduleModelMixin, OnScheduleModelMixin
from edc_visit_tracking.model_mixins import VisitModelMixin

from edc_pharmacy.model_mixins import StudyMedicationCrfModelMixin

from .consents import consent_v1


class SubjectConsent(
    SiteModelMixin,
    NonUniqueSubjectIdentifierFieldMixin,
    UpdatesOrCreatesRegistrationModelMixin,
    BaseUuidModel,
):
    report_datetime = models.DateTimeField(default=get_utcnow)

    consent_datetime = models.DateTimeField(default=get_utcnow)

    version = models.CharField(max_length=25, default="1")

    identity = models.CharField(max_length=25)

    confirm_identity = models.CharField(max_length=25)

    dob = models.DateField(default=date(1995, 1, 1))


class SubjectConsentV1(SubjectConsent):
    objects = ConsentObjectsByCdefManager()
    on_site = CurrentSiteByCdefManager()

    class Meta:
        proxy = True


class SubjectScreening(ScreeningModelMixin, BaseUuidModel):
    consent_definition = consent_v1
    objects = SubjectIdentifierManager()


class SubjectVisit(SiteModelMixin, VisitModelMixin, BaseUuidModel):
    def run_metadata_rules(self, **kwargs):
        pass

    def metadata_create(self):
        pass

    class Meta(VisitModelMixin.Meta, BaseUuidModel.Meta):
        app_label = "edc_pharmacy"


class StudyMedication(
    StudyMedicationCrfModelMixin,
    CrfModelMixin,
    BaseUuidModel,
):
    subject_visit = models.OneToOneField(SubjectVisit, on_delete=models.PROTECT)

    report_datetime = models.DateTimeField(default=get_utcnow)

    def run_metadata_rules_for_related_visit(self, **kwargs):
        pass

    def metadata_update(self, **kwargs):
        pass

    def update_reference_on_save(self):
        return None

    class Meta(CrfModelMixin.Meta, BaseUuidModel.Meta):
        app_label = "edc_pharmacy"


class OnSchedule(SiteModelMixin, OnScheduleModelMixin, BaseUuidModel):
    pass


class OffSchedule(SiteModelMixin, OffScheduleModelMixin, BaseUuidModel):
    pass



# from dateutil.relativedelta import relativedelta
# from django.test import TestCase, override_settings
# from edc_facility import import_holidays
# from edc_list_data import site_list_data
# from edc_randomization.constants import ACTIVE, PLACEBO
# from edc_randomization.site_randomizers import site_randomizers
# from edc_randomization.tests.testcase_mixin import TestCaseMixin, all_sites
# from edc_utils import get_utcnow
#
# from edc_pharmacy.exceptions import InsufficientQuantityError
# from edc_pharmacy.models import (
#     Box,
#     ContainerType,
#     Formulation,
#     FormulationType,
#     Location,
#     Lot,
#     Medication,
#     Product,
#     Room,
#     Route,
#     Shelf,
#     Stock,
#     Units,
#     repackage,
#     repackage_for_subject,
# )
#
#
# @override_settings(SITE_ID=1)
# class TestPackaging(TestCaseMixin, TestCase):
#     import_randomization_list = True
#     site_names = [x.name for x in all_sites]
#     sid_count_for_tests = 5
#
#     @classmethod
#     def setUpTestData(cls):
#         import_holidays(test=True)
#         if cls.import_randomization_list:
#             randomizer_cls = site_randomizers.get("default")
#             randomizer_cls.import_list(
#                 verbose=False, sid_count_for_tests=cls.sid_count_for_tests
#             )
#         site_list_data.initialize()
#         site_list_data.autodiscover()
#
#     def setUp(self):
#         self.medication = Medication.objects.create(
#             name="METFORMIN",
#         )
#
#         self.formulation = Formulation.objects.create(
#             medication=self.medication,
#             strength=500,
#             units=Units.objects.get(name="mg"),
#             route=Route.objects.get(display_name="Oral"),
#             formulation_type=FormulationType.objects.get(display_name__iexact="Tablet"),
#         )
#         self.lot_one = Lot.objects.create(
#             lot_no="LOT1111111",
#             formulation=self.formulation,
#             expiration_date=get_utcnow() + relativedelta(years=1),
#             assignment=ACTIVE,
#         )
#         self.lot_two = Lot.objects.create(
#             lot_no="LOT222222",
#             formulation=self.formulation,
#             expiration_date=get_utcnow() + relativedelta(years=1),
#             assignment=PLACEBO,
#         )
#
#     def test_repack(self):
#         container_type = ContainerType.objects.create(name="bottle")
#
#         location = Location.objects.create(name="LOCATION_ONE")
#         room = Room.objects.create(name="ROOM_ONE", location=location)
#         shelf_one = Shelf.objects.create(name="SHELF_ONE", room=room)
#         shelf_two = Shelf.objects.create(name="SHELF_TWO", room=room)
#         box_two = Box.objects.create(name="BOX_TWO", shelf=shelf_two)
#         box_one = Box.objects.create(name="BOX_ONE", shelf=shelf_one)
#         for i in [1, 2, 3, 4, 5, 6]:
#             Product.objects.create(
#                 product_identifier=f"XYZ{i}",
#                 container_type=container_type,
#                 name=f"BOTTLE_{i}",
#                 lot=self.lot_one,
#                 unit_qty=500,
#             )
#         # package into box
#         product_identifiers = "\n".join(
#             [obj.product_identifier for obj in Product.objects.all()]
#         )
#         Stock.objects.create()
#
#         source_pill_bottle = Product.objects.get(name="XYZ1")
#
#         for _ in [1, 2, 3]:
#             _, source_pill_bottle = repackage(
#                 Product, 128, box=box_two, source_container=source_pill_bottle
#             )
#
#         self.assertRaises(
#             InsufficientQuantityError,
#             repackage,
#             Product,
#             128,
#             box=box_two,
#             source_container=source_pill_bottle,
#         )
#
#         self.assertEqual(
#             source_pill_bottle.unit_qty - source_pill_bottle.unit_qty_out, 500 - (128 * 3)
#         )
#
#         for obj in Product.objects.filter(source_container=source_pill_bottle):
#             self.assertEqual(obj.unit_qty, 128)
#
#     def test_repack_for_subject(self):
#         container_type = ContainerType.objects.create(name="bottle")
#         location = Location.objects.create(name="LOCATION_ONE")
#         room = Room.objects.create(name="ROOM_ONE", location=location)
#         shelf_one = Shelf.objects.create(name="SHELF_ONE", room=room)
#         shelf_two = Shelf.objects.create(name="SHELF_TWO", room=room)
#         box_two = Box.objects.create(name="BOX_TWO", shelf=shelf_two)
#         box_one = Box.objects.create(name="BOX_ONE", shelf=shelf_one)
#         for i in [1, 2, 3, 4, 5, 6]:
#             Product.objects.create(
#                 container_type=container_type,
#                 name=f"BOTTLE_{i}",
#                 lot=self.lot_one,
#                 unit_qty=500,
#                 box=box_one,
#             )
#
#         source_pill_bottle = Product.objects.get(name="BOTTLE_1")
#
#         subject_pill_bottle, source_pill_bottle = repackage_for_subject(
#             new_container_model_cls=Product,
#             unit_qty=32,
#             randomizer_name="default",
#             rando_sid="1001",
#             subject_identifier=None,
#             source_container=source_pill_bottle,
#             box=box_two,
#         )
#
#         self.assertEqual(source_pill_bottle.unit_qty_out, 32)
#         self.assertEqual(subject_pill_bottle.unit_qty, 32)



from __future__ import annotations

from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from edc_utils import get_utcnow

if TYPE_CHECKING:
    from ..models import StockTransfer


def transfer_stock(
    stock_transfer: StockTransfer, stock_codes: list[str], username: str | None = None
) -> tuple[list[str], list[str]]:
    stock_model_cls = django_apps.get_model("edc_pharmacy.stock")
    stock_transfer_item_model_cls = django_apps.get_model("edc_pharmacy.stocktransferitem")
    transferred, skipped_codes = [], []
    for stock_code in stock_codes:
        try:
            stock = stock_model_cls.objects.get(
                code=stock_code,
                allocation__isnull=False,
                confirmed=True,
                location=stock_transfer.from_location,
            )
        except ObjectDoesNotExist:
            skipped_codes.append(stock_code)
        else:
            stock_transfer_item_model_cls.objects.create(
                stock=stock,
                stock_transfer=stock_transfer,
                user_created=username,
                created=get_utcnow(),
            )
            stock.location = stock_transfer.to_location
            stock.save()
            transferred.append(stock_code)
    return transferred, skipped_codes


__all__ = ["transfer_stock"]



from secrets import choice

from dateutil.relativedelta import relativedelta
from django.contrib.sites.models import Site
from django.db.models import Sum
from django.test import TestCase
from edc_consent import site_consents
from edc_constants.constants import COMPLETE, FEMALE, MALE
from edc_list_data import site_list_data
from edc_randomization.constants import ACTIVE, PLACEBO
from edc_randomization.models import RandomizationList
from edc_registration.models import RegisteredSubject
from edc_utils import get_utcnow
from sequences import get_next_value

from ...exceptions import RepackRequestError
from ...models import (
    Assignment,
    Container,
    ContainerType,
    ContainerUnits,
    Formulation,
    FormulationType,
    Location,
    Lot,
    Medication,
    Order,
    OrderItem,
    Product,
    Receive,
    ReceiveItem,
    RepackRequest,
    Route,
    Stock,
    Units,
)
from ...utils import confirm_stock, process_repack_request
from ..consents import consent_v1


class TestOrderReceive(TestCase):
    def setUp(self):
        site_list_data.initialize()
        site_list_data.autodiscover()
        site_consents.registry = {}
        site_consents.loaded = False
        site_consents.register(consent_v1)

        self.medication = Medication.objects.create(
            name="METFORMIN",
        )

        self.formulation = Formulation.objects.create(
            medication=self.medication,
            strength=500,
            units=Units.objects.get(name="mg"),
            route=Route.objects.get(display_name="Oral"),
            formulation_type=FormulationType.objects.get(display_name__iexact="Tablet"),
        )
        self.assignment_active = Assignment.objects.create(name=ACTIVE)
        self.assignment_placebo = Assignment.objects.create(name=PLACEBO)
        self.product_active, self.product_placebo = self.make_products()
        self.lot_active = Lot.objects.create(
            lot_no="1234",
            assignment=self.assignment_active,
            expiration_date=get_utcnow() + relativedelta(years=1),
            product=self.product_active,
        )
        self.lot_placebo = Lot.objects.create(
            lot_no="4321",
            assignment=self.assignment_placebo,
            expiration_date=get_utcnow() + relativedelta(years=1),
            product=self.product_placebo,
        )
        self.location = Location.objects.create(name="central_pharmacy")
        self.location_amana = Location.objects.create(name="amana_pharmacy")

    def make_products(self):
        product_active = Product.objects.create(
            formulation=self.formulation,
            assignment=self.assignment_active,
        )
        product_placebo = Product.objects.create(
            formulation=self.formulation,
            assignment=self.assignment_placebo,
        )
        return product_active, product_placebo

    def make_order(self, container, qty: int | None = None):
        qty = qty or 100
        # product_active, product_placebo = self.make_products()
        order = Order.objects.create(order_datetime=get_utcnow(), item_count=20)
        for i in range(0, 10):
            OrderItem.objects.create(
                order=order,
                product=self.product_active,
                qty=qty,
                container=container,
            )
        for i in range(10, 20):
            OrderItem.objects.create(
                order=order,
                product=self.product_placebo,
                qty=qty,
                container=container,
            )
        order.refresh_from_db()
        return order

    # def test_make_product(self):
    #     self.make_products()

    def test_make_order(self):
        """Test creating an order.

        1. Create products
        2. Create a new order
        3. Add order items to the order for the products
        """
        container_units, _ = ContainerUnits.objects.get_or_create(
            name="tablet", plural_name="tablets"
        )
        container_type, _ = ContainerType.objects.get_or_create(name="tablet")
        container = Container.objects.create(
            container_type=container_type,
            qty=1,
            units=container_units,
            may_order_as=True,
        )
        order = self.make_order(container)
        self.assertEqual(OrderItem.objects.all().count(), 20)
        self.assertEqual(order.item_count, 20)

    def test_receive_ordered_items(self):
        container_units, _ = ContainerUnits.objects.get_or_create(
            name="tablet", plural_name="tablets"
        )
        container_type, _ = ContainerType.objects.get_or_create(name="tablet")
        container = Container.objects.create(
            container_type=container_type,
            qty=1,
            units=container_units,
            may_order_as=True,
            may_receive_as=True,
        )
        order = self.make_order(container)
        receive = Receive.objects.create(order=order, location=self.location)
        order_items = order.orderitem_set.all()
        sums = OrderItem.objects.filter(order=order).aggregate(
            unit_qty=Sum("unit_qty"),
            unit_qty_received=Sum("unit_qty_received"),
        )
        self.assertEqual(sums["unit_qty"], 2000)
        self.assertEqual(sums["unit_qty_received"], None)

        for order_item in order_items:
            obj = ReceiveItem.objects.create(
                receive=receive,
                order_item=order_item,
                qty=100,
                container=container,
                lot=(
                    self.lot_active
                    if order_item.product.assignment.name == "active"
                    else self.lot_placebo
                ),
            )
            # assert container qty received
            self.assertEqual(obj.unit_qty, 100)

        # assert updates order_item.qty_received
        sums = OrderItem.objects.filter(order=order).aggregate(
            unit_qty=Sum("unit_qty"),
            unit_qty_received=Sum("unit_qty_received"),
        )
        self.assertEqual(sums["unit_qty"], 0)
        self.assertEqual(sums["unit_qty_received"], 2000)

        # assert updates order_item.status
        for order_item in order_items:
            self.assertEqual(order_item.status, COMPLETE)

        # assert updates order.status
        order.refresh_from_db()
        self.assertEqual(order.status, COMPLETE)

        # assert added to stock
        self.assertEqual(
            Stock.objects.filter(receive_item__receive=receive).aggregate(
                qty_in=Sum("qty_in")
            )["qty_in"],
            2000,
        )

    def test_receive_ordered_items2(self):
        """Test receive where order product unit (e.g. Tablet) is not
        the same as received product unit (Bottle of 100 tablets).

        That is, we ordered 2000 tablets and received 20 bottles
        of 100 tablets
        """
        # order 2000 tablets
        container_units, _ = ContainerUnits.objects.get_or_create(
            name="tablet", plural_name="tablets"
        )
        container_type, _ = ContainerType.objects.get_or_create(name="tablet")
        container_2000 = Container.objects.create(
            container_type=container_type,
            qty=1,
            units=container_units,
            may_order_as=True,
        )
        order = self.make_order(container_2000)

        # receive 20 bottles or 100
        container_type, _ = ContainerType.objects.get_or_create(name="bottle")
        container_20 = Container.objects.create(
            container_type=container_type,
            qty=100,
            units=container_units,
            may_receive_as=True,
        )

        receive = Receive.objects.create(order=order, location=self.location)
        order_items = order.orderitem_set.all()
        for order_item in order_items:
            ReceiveItem.objects.create(
                receive=receive,
                order_item=order_item,
                qty=1,
                container=container_20,
                lot=(
                    self.lot_active
                    if order_item.product.assignment.name == "active"
                    else self.lot_placebo
                ),
            )

        # assert updates order_item.qty_received
        sums = OrderItem.objects.filter(order=order).aggregate(
            unit_qty=Sum("unit_qty"), unit_qty_received=Sum("unit_qty_received")
        )
        self.assertEqual(sums["unit_qty"], 0)
        self.assertEqual(sums["unit_qty_received"], 2000)

        # assert updates order_item.status
        for order_item in order_items:
            self.assertEqual(order_item.status, COMPLETE)

        # assert updates order.status
        order.refresh_from_db()
        self.assertEqual(order.status, COMPLETE)

        # assert added to stock
        self.assertEqual(
            Stock.objects.filter(receive_item__receive=receive).aggregate(
                qty_in=Sum("qty_in")
            )["qty_in"],
            20,
        )
        self.assertEqual(
            Stock.objects.filter(receive_item__receive=receive).aggregate(
                unit_qty_in=Sum("unit_qty_in")
            )["unit_qty_in"],
            2000,
        )

    def order_and_receive(self):
        # product_active, product_placebo = self.make_products()
        container_units, _ = ContainerUnits.objects.get_or_create(
            name="tablet", plural_name="tablets"
        )
        container_type, _ = ContainerType.objects.get_or_create(name="tablet")
        container = Container.objects.create(
            container_type=container_type,
            qty=1,
            units=container_units,
            may_order_as=True,
        )
        order = Order.objects.create(order_datetime=get_utcnow())
        OrderItem.objects.create(
            order=order,
            product=self.product_active,
            qty=50000,
            container=container,
        )
        OrderItem.objects.create(
            order=order,
            product=self.product_placebo,
            qty=50000,
            container=container,
        )
        order.refresh_from_db()

        container_type, _ = ContainerType.objects.get_or_create(name="bottle")
        container_bulk = Container.objects.create(
            container_type=container_type,
            qty=5000,
            units=container_units,
            may_receive_as=True,
        )

        receive = Receive.objects.create(order=order, location=self.location)
        order_items = order.orderitem_set.all()
        for order_item in order_items:
            ReceiveItem.objects.create(
                receive=receive,
                order_item=order_item,
                qty=10,
                container=container_bulk,
                lot=(
                    self.lot_active
                    if order_item.product.assignment.name == "active"
                    else self.lot_placebo
                ),
            )
        receive.save()
        return receive

    def test_delete_receive_item(self):
        # confirm deleting stock, resave received items recreates
        self.order_and_receive()
        Stock.objects.all().delete()
        for obj in ReceiveItem.objects.all():
            obj.save()
        # assert stock count is equal to receiveditem.qty *
        # receiveitem.count()
        self.assertEqual(
            Stock.objects.all().count(),
            int(ReceiveItem.objects.values("qty").aggregate(qty=Sum("qty")).get("qty")),
        )

        # confirm deleting stock & received items resets unit_qty_received on order items
        Stock.objects.all().delete()
        ReceiveItem.objects.all().delete()
        for order_item in OrderItem.objects.all():
            self.assertEqual(0, order_item.unit_qty_received)

    def get_container_5000(self) -> Container:
        container_units, _ = ContainerUnits.objects.get_or_create(
            name="tablet", plural_name="tablets"
        )
        container_type, _ = ContainerType.objects.get_or_create(name="bottle")
        container_5000, _ = Container.objects.get_or_create(
            qty=5000, container_type=container_type
        )
        return container_5000

    def get_container_128(self) -> Container:
        container_units, _ = ContainerUnits.objects.get_or_create(
            name="tablet", plural_name="tablets"
        )
        container_type, _ = ContainerType.objects.get_or_create(name="bottle")
        container_128, _ = Container.objects.get_or_create(
            qty=128, container_type=container_type
        )
        return container_128

    def test_repack(self):
        """Test repackage two bottles of 50000 into bottles of 128."""
        # create order of 50000 for each arm
        receive = self.order_and_receive()

        # assert created stock
        container_5000 = self.get_container_5000()
        container_128 = self.get_container_128()

        # assert
        self.assertEqual(Stock.objects.filter(container=container_5000).count(), 20)
        self.assertEqual(
            Stock.objects.values("container__name")
            .annotate(count=Sum("qty_in"))[0]
            .get("count"),
            20,
        )
        self.assertEqual(
            Stock.objects.values("container__name")
            .annotate(count=Sum("unit_qty_in"))[0]
            .get("count"),
            100000,
        )

        # assert raises if stock not confirmed
        stock = Stock.objects.filter(container=container_5000).first()
        with self.assertRaises(RepackRequestError) as cm:
            repack_request = RepackRequest.objects.create(
                from_stock=stock,
                container=container_128,
                requested_qty=39,
            )
        self.assertIn("Unconfirmed stock item", str(cm.exception))

        # confirm stock items
        for stock in Stock.objects.filter(container=container_5000):
            confirm_stock(receive, [stock.code], fk_attr="receive_item__receive")

        # REPACK REQUEST **********************************************
        for stock in Stock.objects.filter(container=container_5000):
            # create request
            repack_request = RepackRequest.objects.create(
                from_stock=stock,
                container=container_128,
                requested_qty=39,
            )
            # process
            process_repack_request(repack_request.pk, username=None)

        for repack_request in RepackRequest.objects.all():
            # assert unconfirmed stock instances (central)
            self.assertEqual(repack_request.stock_set.filter(confirmed=False).count(), 39)

        # scan in some or all labels to confirm stock (central)
        for repack_request in RepackRequest.objects.all():
            codes = [obj.code for obj in repack_request.stock_set.filter(confirmed=False)]
            confirmed, already_confirmed, invalid = confirm_stock(
                repack_request, codes, fk_attr="repack_request"
            )

            self.assertEqual(len(confirmed), 39)
            self.assertEqual(len(already_confirmed), 0)
            self.assertEqual(len(invalid), 0)

        for repack_request in RepackRequest.objects.all():
            codes = [
                f"{obj.code}blah" for obj in repack_request.stock_set.filter(confirmed=True)
            ]
            confirmed, already_confirmed, invalid = confirm_stock(
                repack_request, codes, fk_attr="repack_request"
            )

            self.assertEqual(len(confirmed), 0)
            self.assertEqual(len(already_confirmed), 0)
            self.assertEqual(len(invalid), 39)

        for repack_request in RepackRequest.objects.all():
            codes = [obj.code for obj in repack_request.stock_set.filter(confirmed=True)]
            confirmed, already_confirmed, invalid = confirm_stock(
                repack_request, codes, fk_attr="repack_request"
            )

            self.assertEqual(len(confirmed), 0)
            self.assertEqual(len(already_confirmed), 39)
            self.assertEqual(len(invalid), 0)

        for repack_request in RepackRequest.objects.all():
            # assert unconfirmed stock instances (central)
            self.assertEqual(repack_request.stock_set.filter(confirmed=False).count(), 0)
            # assert confirmed stock instances (central)
            self.assertEqual(repack_request.stock_set.filter(confirmed=True).count(), 39)

        # refer back to repack_request from stock
        self.assertEqual(Stock.objects.filter(repack_request__isnull=True).count(), 20)
        self.assertEqual(Stock.objects.filter(repack_request__isnull=False).count(), 39 * 20)

        # we create bottles of 128 from the two bottles of 50000 tablets
        # the total number of tablets remains the same
        unit_qty_in = Stock.objects.all().aggregate(unit_qty_in=Sum("unit_qty_in"))[
            "unit_qty_in"
        ]
        self.assertEqual(unit_qty_in, 199840)

        unit_qty_out = Stock.objects.all().aggregate(unit_qty_out=Sum("unit_qty_out"))[
            "unit_qty_out"
        ]
        self.assertEqual(unit_qty_out, 99840)

        # repackaged 99840 tablets into bottles of 128
        unit_qty_in = Stock.objects.filter(container=container_128).aggregate(
            unit_qty_in=Sum("unit_qty_in")
        )["unit_qty_in"]
        self.assertEqual(unit_qty_in, 99840)

        unit_qty_out = Stock.objects.filter(container=container_128).aggregate(
            unit_qty_out=Sum("unit_qty_out")
        )["unit_qty_out"]
        self.assertEqual(unit_qty_out, 0)

        # 160 tablets leftover in the two bottles with capacity of 50000
        unit_qty_in = Stock.objects.filter(container=container_5000).aggregate(
            unit_qty_in=Sum("unit_qty_in")
        )["unit_qty_in"]
        self.assertEqual(unit_qty_in, 100000)
        unit_qty_out = Stock.objects.filter(container=container_5000).aggregate(
            unit_qty_out=Sum("unit_qty_out")
        )["unit_qty_out"]
        self.assertEqual(unit_qty_out, 99840)
        self.assertEqual(unit_qty_in - unit_qty_out, 160)

    def make_randomized_subject(self, site: Site, subject_count: int):
        subjects = {}
        for i in range(0, subject_count):
            subjects.update({f"S{i:4d}": choice([self.product_placebo, self.product_active])})
        for subject_identifier, product in subjects.items():
            RegisteredSubject.objects.create(
                subject_identifier=subject_identifier,
                gender=choice([MALE, FEMALE]),
                randomization_list_model="edc_randomization.randomizationlist",
            )
            RandomizationList.objects.create(
                sid=get_next_value(sequence_name=RandomizationList._meta.label_lower),
                subject_identifier=subject_identifier,
                assignment=product.assignment.name,
                allocated_site=site,
                site_name=site.name,
                allocated=True,
                allocated_datetime=get_utcnow(),
            )
        return subjects

    def test_allocate_to_subject(self):
        self.order_and_receive()



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from edc_utils import get_utcnow

if TYPE_CHECKING:
    from ..models import Receive, RepackRequest


def confirm_stock(
    obj: RepackRequest | Receive | None,
    stock_codes: list[str],
    fk_attr: str | None = None,
    confirmed_by: str | None = None,
    user_created: str = None,
    created: datetime = None,
) -> tuple[list[str], list[str], list[str]]:
    """Confirm stock instances given a list of stock codes
    and a request/receive pk.

    Called from ConfirmStock view.

    See also: confirm_stock_action
    """
    stock_model_cls = django_apps.get_model("edc_pharmacy.stock")
    stock_codes = [s.strip() for s in stock_codes]
    invalid = []
    confirmed = []
    already_confirmed = []
    opts = {}
    if obj:
        opts = {fk_attr: obj.id}
    for stock_code in stock_codes:
        try:
            stock = stock_model_cls.objects.get(
                code=stock_code,
                **opts,
            )
        except ObjectDoesNotExist:
            invalid.append(stock_code)
        else:
            if not stock.confirmed:
                stock.confirmed = True
                stock.user_modified = user_created
                stock.modified = created
                stock.confirmed_datetime = get_utcnow()
                stock.confirmed_by = confirmed_by or user_created
                stock.save(
                    update_fields=[
                        "confirmed",
                        "confirmed_datetime",
                        "confirmed_by",
                        "user_modified",
                        "modified",
                    ]
                )
                confirmed.append(stock.code)
            else:
                already_confirmed.append(stock.code)
    return confirmed, already_confirmed, invalid


__all__ = ["confirm_stock"]



# Generated by Django 5.1.2 on 2024-11-18 00:45

import _socket
import django.db.models.deletion
import django_audit_fields.fields.hostname_modification_field
import django_audit_fields.fields.userfield
import django_audit_fields.fields.uuid_auto_field
import django_audit_fields.models.audit_model_mixin
import django_revision.revision_field
import edc_pharmacy.models.stock.dispense_item
import edc_utils.date
import simple_history.models
import uuid
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        (
            "edc_pharmacy",
            "0038_historicalstocktransferconfirmation_transfer_confirmation_identifier_and_more",
        ),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.RemoveField(
            model_name="dispense",
            name="registered_subject",
        ),
        migrations.RemoveField(
            model_name="historicaldispense",
            name="registered_subject",
        ),
        migrations.AddField(
            model_name="dispense",
            name="dispensed_by",
            field=models.CharField(blank=True, max_length=100, null=True),
        ),
        migrations.AddField(
            model_name="dispense",
            name="rx",
            field=models.ForeignKey(
                null=True, on_delete=django.db.models.deletion.PROTECT, to="edc_pharmacy.rx"
            ),
        ),
        migrations.AddField(
            model_name="historicaldispense",
            name="dispensed_by",
            field=models.CharField(blank=True, max_length=100, null=True),
        ),
        migrations.AddField(
            model_name="historicaldispense",
            name="rx",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="edc_pharmacy.rx",
            ),
        ),
        migrations.AlterField(
            model_name="stocktransferconfirmation",
            name="stock",
            field=models.OneToOneField(
                on_delete=django.db.models.deletion.PROTECT, to="edc_pharmacy.stock"
            ),
        ),
        migrations.CreateModel(
            name="HistoricalDispenseItem",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True, default=django_audit_fields.models.audit_model_mixin.utcnow
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True, default=django_audit_fields.models.audit_model_mixin.utcnow
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                        verbose_name="Hostname created",
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                        verbose_name="Hostname modified",
                    ),
                ),
                (
                    "device_created",
                    models.CharField(blank=True, max_length=10, verbose_name="Device created"),
                ),
                (
                    "device_modified",
                    models.CharField(
                        blank=True, max_length=10, verbose_name="Device modified"
                    ),
                ),
                (
                    "locale_created",
                    models.CharField(
                        blank=True,
                        help_text="Auto-updated by Modeladmin",
                        max_length=10,
                        null=True,
                        verbose_name="Locale created",
                    ),
                ),
                (
                    "locale_modified",
                    models.CharField(
                        blank=True,
                        help_text="Auto-updated by Modeladmin",
                        max_length=10,
                        null=True,
                        verbose_name="Locale modified",
                    ),
                ),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        db_index=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                    ),
                ),
                ("visit_code", models.CharField(editable=False, max_length=25, null=True)),
                (
                    "visit_code_sequence",
                    models.IntegerField(
                        blank=True,
                        default=0,
                        help_text="An integer to represent the sequence of additional appointments relative to the base appointment, 0, needed to complete data collection for the timepoint. (NNNN.0)",
                        null=True,
                        verbose_name="Sequence",
                    ),
                ),
                (
                    "dispense_item_identifier",
                    models.CharField(
                        blank=True,
                        db_index=True,
                        help_text="A sequential unique identifier set by the EDC",
                        max_length=36,
                        null=True,
                    ),
                ),
                (
                    "dispense_item_datetime",
                    models.DateTimeField(default=edc_utils.date.get_utcnow),
                ),
                (
                    "history_id",
                    models.UUIDField(
                        default=uuid.uuid4, editable=False, primary_key=True, serialize=False
                    ),
                ),
                ("history_date", models.DateTimeField(db_index=True)),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "dispense",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="edc_pharmacy.dispense",
                        verbose_name="Stock request",
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "stock",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="edc_pharmacy.stock",
                    ),
                ),
            ],
            options={
                "verbose_name": "historical Dispense item",
                "verbose_name_plural": "historical Dispense items",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": ("history_date", "history_id"),
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="DispenseItem",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True, default=django_audit_fields.models.audit_model_mixin.utcnow
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True, default=django_audit_fields.models.audit_model_mixin.utcnow
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                        verbose_name="Hostname created",
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                        verbose_name="Hostname modified",
                    ),
                ),
                (
                    "device_created",
                    models.CharField(blank=True, max_length=10, verbose_name="Device created"),
                ),
                (
                    "device_modified",
                    models.CharField(
                        blank=True, max_length=10, verbose_name="Device modified"
                    ),
                ),
                (
                    "locale_created",
                    models.CharField(
                        blank=True,
                        help_text="Auto-updated by Modeladmin",
                        max_length=10,
                        null=True,
                        verbose_name="Locale created",
                    ),
                ),
                (
                    "locale_modified",
                    models.CharField(
                        blank=True,
                        help_text="Auto-updated by Modeladmin",
                        max_length=10,
                        null=True,
                        verbose_name="Locale modified",
                    ),
                ),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("visit_code", models.CharField(editable=False, max_length=25, null=True)),
                (
                    "visit_code_sequence",
                    models.IntegerField(
                        blank=True,
                        default=0,
                        help_text="An integer to represent the sequence of additional appointments relative to the base appointment, 0, needed to complete data collection for the timepoint. (NNNN.0)",
                        null=True,
                        verbose_name="Sequence",
                    ),
                ),
                (
                    "dispense_item_identifier",
                    models.CharField(
                        blank=True,
                        help_text="A sequential unique identifier set by the EDC",
                        max_length=36,
                        null=True,
                        unique=True,
                    ),
                ),
                (
                    "dispense_item_datetime",
                    models.DateTimeField(default=edc_utils.date.get_utcnow),
                ),
                (
                    "dispense",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="edc_pharmacy.dispense",
                        verbose_name="Stock request",
                    ),
                ),
                (
                    "stock",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.PROTECT, to="edc_pharmacy.stock"
                    ),
                ),
            ],
            options={
                "verbose_name": "Dispense item",
                "verbose_name_plural": "Dispense items",
                "abstract": False,
                "default_permissions": ("add", "change", "delete", "view", "export", "import"),
                "default_manager_name": "objects",
                "indexes": [
                    models.Index(
                        fields=["modified", "created"], name="edc_pharmac_modifie_989071_idx"
                    ),
                    models.Index(
                        fields=["user_modified", "user_created"],
                        name="edc_pharmac_user_mo_95fc80_idx",
                    ),
                ],
            },
            managers=[
                ("objects", edc_pharmacy.models.stock.dispense_item.Manager()),
            ],
        ),
    ]



from __future__ import annotations

from decimal import Decimal

from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.db.models import PROTECT
from edc_model.models import BaseUuidModel, HistoricalRecords
from edc_utils import get_utcnow
from sequences import get_next_value

from ...choices import STOCK_STATUS
from ...constants import ALLOCATED, AVAILABLE, ZERO_ITEM
from ...exceptions import AllocationError, AssignmentError
from ...utils import get_random_code
from .allocation import Allocation
from .container import Container
from .dispense import Dispense
from .location import Location
from .lot import Lot
from .managers import StockManager
from .product import Product
from .receive_item import ReceiveItem
from .repack_request import RepackRequest


class Stock(BaseUuidModel):

    stock_identifier = models.CharField(
        verbose_name="Internal stock identifier",
        max_length=36,
        unique=True,
        null=True,
        blank=True,
        help_text="A sequential unique identifier set by the EDC",
    )

    code = models.CharField(
        verbose_name="Stock code",
        max_length=15,
        unique=True,
        null=True,
        blank=True,
        help_text="A unique alphanumeric code",
    )

    stock_datetime = models.DateTimeField(
        default=get_utcnow, help_text="date stock record created"
    )

    receive_item = models.ForeignKey(
        ReceiveItem, on_delete=models.PROTECT, null=True, blank=False
    )

    repack_request = models.ForeignKey(
        RepackRequest, on_delete=models.PROTECT, null=True, blank=True
    )

    from_stock = models.ForeignKey(
        "edc_pharmacy.stock",
        related_name="source_stock",
        on_delete=models.PROTECT,
        null=True,
    )

    allocation = models.OneToOneField(
        Allocation,
        on_delete=models.PROTECT,
        null=True,
        blank=True,
        help_text="Subject allocation",
    )

    dispense = models.OneToOneField(
        Dispense,
        verbose_name="Dispense",
        on_delete=models.PROTECT,
        null=True,
        blank=True,
        help_text="Stock was dispensed",
    )

    product = models.ForeignKey(Product, on_delete=models.PROTECT)

    lot = models.ForeignKey(
        Lot, verbose_name="Batch", on_delete=models.PROTECT, null=True, blank=False
    )

    container = models.ForeignKey(Container, on_delete=models.PROTECT, null=True, blank=False)

    location = models.ForeignKey(Location, on_delete=PROTECT, null=True, blank=False)

    qty = models.DecimalField(
        null=True,
        blank=False,
        decimal_places=2,
        max_digits=20,
        default=Decimal(0.0),
    )

    qty_in = models.DecimalField(
        null=True,
        blank=False,
        decimal_places=2,
        max_digits=20,
        default=Decimal(0.0),
        validators=[MinValueValidator(0), MaxValueValidator(1)],
    )

    qty_out = models.DecimalField(
        decimal_places=2,
        max_digits=20,
        default=Decimal(0.0),
        validators=[MinValueValidator(0), MaxValueValidator(1)],
    )

    unit_qty_in = models.DecimalField(
        decimal_places=2,
        max_digits=20,
        default=Decimal(0.0),
        validators=[MinValueValidator(0)],
    )

    unit_qty_out = models.DecimalField(
        decimal_places=2,
        max_digits=20,
        default=Decimal(0.0),
        validators=[MinValueValidator(0)],
    )

    status = models.CharField(max_length=25, choices=STOCK_STATUS, default=AVAILABLE)

    description = models.CharField(max_length=100, null=True, blank=True)

    allocated = models.BooleanField(default=False)

    transferred = models.BooleanField(default=False)

    confirmed = models.BooleanField(
        default=False,
        help_text=(
            "True if stock was labelled and confirmed; "
            "False if stock was received/repacked but never confirmed."
        ),
    )
    confirmed_datetime = models.DateTimeField(null=True, blank=True)
    confirmed_by = models.CharField(max_length=150, null=True, blank=True)

    confirmed_at_site = models.BooleanField(default=False)

    stored_at_site = models.BooleanField(default=False)

    dispensed = models.BooleanField(default=False)

    allocated_datetime = models.DateTimeField(null=True, blank=True)

    subject_identifier = models.CharField(max_length=50, null=True, blank=True)

    objects = StockManager()

    history = HistoricalRecords()

    def __str__(self):
        return f"{self.code}: {self.description}"

    def save(self, *args, **kwargs):
        if not self.stock_identifier:
            next_id = get_next_value(self._meta.label_lower)
            self.stock_identifier = f"{next_id:010d}"
            self.code = get_random_code(self.__class__, 6, 10000)
            self.product = self.get_receive_item().order_item.product
        if not self.description:
            self.description = f"{self.product.name} - {self.container.name}"
        self.verify_assignment_or_raise()
        self.verify_assignment_or_raise(self.from_stock)
        self.update_status()
        self.unit_qty_in = Decimal(self.qty_in) * Decimal(self.container.qty)
        super().save(*args, **kwargs)

    def update_transferred(self) -> bool:
        if (
            self.allocation
            and self.allocation.stock_request_item.stock_request.location == self.location
            and self.container.may_request_as
        ):
            return True
        return False

    def verify_assignment_or_raise(
        self, stock: models.ForeignKey[Stock] | None = None
    ) -> None:
        """Verify that the LOT and PRODUCT assignments match."""
        if not stock:
            stock = self
        if stock.product.assignment != stock.lot.assignment:
            raise AssignmentError("Lot number assignment does not match product assignment!")
        if self.allocation and self.allocation.assignment != stock.lot.assignment:
            raise AllocationError(
                f"Allocation assignment does not match lot assignment! Got {self.code}."
            )

    def update_status(self):
        if self.allocation:
            self.status = ALLOCATED
        elif self.qty_out == self.qty_in:
            self.status = ZERO_ITEM
        else:
            self.status = AVAILABLE

    def get_receive_item(self) -> ReceiveItem:
        """Recursively fetch the original receive item."""
        obj: Stock = self
        receive_item = self.receive_item
        while not receive_item:
            obj = obj.from_stock  # noqa
            receive_item = obj.receive_item
        return receive_item

    @property
    def unit_qty(self):
        """Unit qty of hand"""
        return self.unit_qty_in - self.unit_qty_out

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Stock"
        verbose_name_plural = "Stock"



from django import forms

from ...models import StockTransfer


class StockTransferForm(forms.ModelForm):

    def clean(self):
        cleaned_data = super().clean()

        return cleaned_data

    class Meta:
        model = StockTransfer
        fields = "__all__"
        help_text = {"transfer_identifier": "(read-only)"}
        widgets = {
            "transfer_identifier": forms.TextInput(attrs={"readonly": "readonly"}),
        }



from django.contrib import admin
from edc_list_data.admin import ListModelAdminMixin

from ...admin_site import edc_pharmacy_admin
from ...forms import ContainerTypeForm
from ...models import ContainerType


@admin.register(ContainerType, site=edc_pharmacy_admin)
class ContainerTypeAdmin(ListModelAdminMixin, admin.ModelAdmin):
    show_object_tools = True

    form = ContainerTypeForm



from __future__ import annotations

from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.db.models import Count
from django_pandas.io import read_frame

if TYPE_CHECKING:
    import pandas as pd

    from ..models import Location


def get_stock_for_location_df(location: Location) -> pd.DataFrame:
    """Returns a dataframe of all stock records for this
    location.
    """
    stock_model_cls = django_apps.get_model("edc_pharmacy.Stock")
    qs_stock = (
        stock_model_cls.objects.values(
            "allocation__registered_subject__subject_identifier", "code", "dispensed"
        )
        .filter(location=location, qty=1)
        .annotate(count=Count("allocation__registered_subject__subject_identifier"))
    )
    df_stock = read_frame(qs_stock).rename(
        columns={
            "allocation__registered_subject__subject_identifier": "subject_identifier",
            "count": "stock_qty",
        }
    )
    df_stock["dispensed"] = df_stock["dispensed"].astype("boolean").fillna(False)
    return df_stock
